/* FILE: mesh.cc                -*-Mode: c++-*-
 *
 * Abstract mesh class and children
 *
 */

#include <math.h>
#include <stdio.h>
#include <string.h>

#include "nb.h"

#include "fileio.h"
#include "mesh.h"

/* End includes */     // This is an optional directive to build.tcl,
                       // that excludes the remainder of the file from
                       // '#include ".*"' dependency building

//////////////////////////////////////////////////////////////////////////
// Abstract Mesh class
void Vf_Mesh::SetFilename(const char* _Filename) {
  if(Filename!=NULL) delete[] Filename;
  if(_Filename==NULL) Filename=NULL;
  else {
    Filename=new char[strlen(_Filename)+1];
    strcpy(Filename,_Filename);
  }
}
void Vf_Mesh::SetTitle(const char* _Title) {
  if(Title!=NULL) delete[] Title;
  if(_Title==NULL) Title=NULL;
  else {
    Title=new char[strlen(_Title)+1];
    strcpy(Title,_Title);
  }
}
void Vf_Mesh::SetDescription(const char* _Description) {
  if(Description!=NULL) delete[] Description;
  if(_Description==NULL) Description=NULL;
  else {
    Description=new char[strlen(_Description)+1];
    strcpy(Description,_Description);
  }
}
void Vf_Mesh::SetMeshUnit(const char* _MeshUnit) {
  if(MeshUnit!=NULL) delete[] MeshUnit;
  if(_MeshUnit==NULL) MeshUnit=NULL;
  else {
    MeshUnit=new char[strlen(_MeshUnit)+1];
    strcpy(MeshUnit,_MeshUnit);
  }
}
void Vf_Mesh::SetValueUnit(const char* _ValueUnit) {
  if(ValueUnit!=NULL) delete[] ValueUnit;
  if(_ValueUnit==NULL) ValueUnit=NULL;
  else {
    ValueUnit=new char[strlen(_ValueUnit)+1];
    strcpy(ValueUnit,_ValueUnit);
  }
}
const char* Vf_Mesh::GetName() const { 
  if(Filename==NULL) return "";
  return Filename; 
}
const char* Vf_Mesh::GetTitle() const {
  if(Title==NULL) return "";
  return Title; 
}
const char* Vf_Mesh::GetDescription() const {
  if(Description==NULL) return "";
  return Description; 
}
const char* Vf_Mesh::GetMeshUnit() const {
  if(MeshUnit==NULL) return "";
  return MeshUnit; 
}
const char* Vf_Mesh::GetValueUnit() const {
  if(ValueUnit==NULL) return "";
  return ValueUnit; 
}

INT4m Vf_Mesh::GetRange(Nb_BoundingBox<REAL4> &range) const
{
  Nb_BoundingBox<REAL8> wrange;
  INT4m code=GetPreciseRange(wrange);
  Convert(wrange,range);
  return code;
}

INT4m Vf_Mesh::GetDataRange(Nb_BoundingBox<REAL4> &range) const
{
  Nb_BoundingBox<REAL8> wrange;
  INT4m code=GetPreciseDataRange(wrange);
  Convert(wrange,range);
  return code;
}

INT4m Vf_Mesh::FindClosest(const Nb_Vec3<REAL4> &pos,
			   Nb_LocatedVector<REAL4> &lv)
{ // Low-precision version.
  Nb_Vec3<REAL8> spos;
  Nb_LocatedVector<REAL8> slv;
  Convert(pos,spos);
  INT4m code=FindPreciseClosest(spos,slv);
  Convert(slv.location,lv.location);
  Convert(slv.value,lv.value);
  return code;
}

//////////////////////////////////////////////////////////////////////////
// Regular 3D grid of Nb_Vec3<REAL8>'s
const ClassDoc Vf_GridVec3f::class_doc("Vf_GridVec3f",
		    "Michael J. Donahue (michael.donahue@nist.gov)",
		    "1.0.0","11-Dec-1996");

// The following constructor should be called from file input functions.
Vf_GridVec3f::Vf_GridVec3f(const char* filename,
                           const char* title,const char* desc,
			   const char* meshunit,
			   const char* valueunit,REAL8 valuemultiplier,
                           INT4m newi,INT4m newj,INT4m newk,
                           const Nb_Vec3<REAL8> &basept,
                           const Nb_Vec3<REAL8> &gridstep,
                           const Nb_List< Nb_Vec3<REAL8> > &newboundary,
                           const Nb_BoundingBox<REAL8> &newdatarange,
                           const Nb_BoundingBox<REAL8> &newrange)
  :Vf_Mesh("Vf_GridVec3f",filename,title,desc,meshunit,
	   valueunit,valuemultiplier)
{
  SetSize(newi,newj,newk);
  coords_base=basept;
  coords_step=gridstep;
  if(coords_step.x==0.) coords_step.x=1.;
  if(coords_step.y==0.) coords_step.y=1.;
  if(coords_step.z==0.) coords_step.z=1.;
  boundary=newboundary;
  data_range=newdatarange;
  range=newrange;
}

// Problem <-> Nodal Coordinates Conversion Routines
// The conversion from nodes to problem coordinates is
//    (i,j,k) -> base+(i*step.x,j*step.y,k*step.z).
// NOTE 1: The grid vector access functions, e.g., operator(), access
//   via the integral "nodal" coordinates, *not* in terms of the
//   (real valued) problem coordinates.
// NOTE 2: The values in coords_step are guaranteed to be !=0,
//   but may be negative!  However, GetStdStep() (below), returns
//   a value >0.
// NOTE 3: The coords conversion routines don't preserve order
//   if step.? is negative, e.g., say (min.x,min.y) < (max.x,max.y)
//   going in, but step.x is negative.  Then after conversion,
//   min.y~ will still be < max.y~, but now min.x~>max.x~.

Nb_Vec3<REAL8>
Vf_GridVec3f::NodalToProblemCoords(Nb_Vec3<REAL8> const &w_nodal)
  const
{
  Nb_Vec3<REAL8> v(w_nodal);
  v.x*=coords_step.x;
  v.y*=coords_step.y;
  v.z*=coords_step.z;
  v+=coords_base;
  return v;
}

Nb_Vec3<REAL8>
Vf_GridVec3f::ProblemToNodalCoords(Nb_Vec3<REAL8> const &v_problem)
  const
{
  Nb_Vec3<REAL8> w(v_problem);
  w-=coords_base;
  w.x/=coords_step.x;
  w.y/=coords_step.y;
  w.z/=coords_step.z;
  return w;
}

REAL8m
Vf_GridVec3f::GetMaxMag() const
{
  INT4m i,j,k,isize,jsize,ksize;
  GetDimens(isize,jsize,ksize);
  REAL8m magsq,maxmagsq=0.0;
  for(k=0;k<ksize;k++) for(j=0;j<jsize;j++) for(i=0;i<isize;i++) {
    magsq=GridVec(i,j,k).MagSq();
    if(magsq>maxmagsq) maxmagsq=magsq;
  }
  return sqrt(maxmagsq);
}


INT4m
Vf_GridVec3f::GetPreciseRange(Nb_BoundingBox<REAL8> &myrange) const
{
  myrange=range;
  return 0;
}

INT4m
Vf_GridVec3f::GetPreciseDataRange(Nb_BoundingBox<REAL8> &myrange) const
{
  myrange=data_range;
  return 0;
}

INT4m
Vf_GridVec3f::GetBoundaryList(Nb_List< Nb_Vec3<REAL4> > &boundary_list) const
{
  Nb_Vec3<REAL4> newpt;
  const Nb_Vec3<REAL8> *vptr;
  Nb_List_Index< Nb_Vec3<REAL8> > key;
  boundary_list.Clear();
  for(vptr=boundary.GetFirst(key);vptr!=NULL;vptr=boundary.GetNext(key)) {
    newpt.Set(vptr->x,vptr->y,vptr->z);
    boundary_list.Append(newpt);
  }
  return 0;
}

INT4m Vf_GridVec3f::ColorQuantityTypes(Nb_List<Nb_DString> &types) const
{
  types.Clear();
  types.Append(Nb_DString("x"));
  types.Append(Nb_DString("y"));
  types.Append(Nb_DString("z"));
  types.Append(Nb_DString("mag"));
  types.Append(Nb_DString("xy-angle"));
  types.Append(Nb_DString("div"));
  types.Append(Nb_DString("none"));
  return types.GetSize();
}

REAL4m Vf_GridVec3f::GetVecShade(const char* colorquantity,
                                 const Nb_Vec3<REAL4>& v) const
{
  REAL4m shade = 0.5;
  if(strcmp("xy-angle",colorquantity)==0) {
    shade = atan2(v.y,v.x)*(1./(2*PI));
    if(shade<0) shade+=1.0;
  } else {
    REAL8m mag=v.Mag();
    if(UseMagHints) {
      if(mag>MaxMagHint) mag=MaxMagHint;
      if(mag<MinMagHint) mag=1.0;
    }
    REAL8m val=0.0;
    if(strcmp("x",colorquantity)==0)        val = v.x;
    else if(strcmp("y",colorquantity)==0)   val = v.y;
    else if(strcmp("z",colorquantity)==0)   val = v.z;
    else                                    return -1;
    if(fabs(val)<mag) shade = (1.+val/mag)/2.;
    else if(val>0.)   shade = 1.0;
    else              shade = 0.0;
  }
  return shade;
}

INT4m Vf_GridVec3f::FindPreciseClosest(const Nb_Vec3<REAL8> &pos,
				       Nb_LocatedVector<REAL8> &lv)
{
  Nb_Vec3<REAL8> npos=ProblemToNodalCoords(pos);
  INT4m i,isize,j,jsize,k,ksize;
  isize=jsize=ksize=1;  // Just to shut up some compilers.
  GetDimens(isize,jsize,ksize);
  i=INT4m(OC_ROUND(npos.x));
  if(i<0) i=0;  else if(i>=isize) i=isize-1;
  j=INT4m(OC_ROUND(npos.y));
  if(j<0) j=0;  else if(j>=jsize) j=jsize-1;
  k=INT4m(OC_ROUND(npos.z));
  if(k<0) k=0;  else if(k>=ksize) k=ksize-1;

  lv.value=GridVec(i,j,k);
  lv.location=NodalToProblemCoords(Nb_Vec3<REAL8>(REAL8(i),
						  REAL8(j),REAL8(k)));
  return 0;
}

INT4m Vf_GridVec3f::GetDisplayList(REAL4m &xstep_request,
                                   REAL4m &ystep_request,
                                   REAL4m &zstep_request,
                                   const Nb_BoundingBox<REAL4> &dsprange,
                                   const char *colorquantity,
                                   Nb_List<Vf_DisplayVector> &display_list)
{ // Fills display_list with points from mesh inside "dsprange", with
  // step size close to ?step_request, which is reset to value actually
  // used.
  //   Use ?step_request=0 to remove constraint on that axis, e.g.,
  // if xstep_request=ystep_request=zstep_request=0, then this
  // routine places all points inside "dsprange" into display_list.
  //   NOTE: Non-integral values for step_request lead to unattractive
  // banding in the output display list, and so are currently
  // overridden.
#define MEMBERNAME "GetDisplayList"
  if(xstep_request<0 || ystep_request<0 || zstep_request<0)
    FatalError(-1,STDDOC,"%s: xstep=%f, ystep=%f, zstep=%d",
	       ErrBadParam,xstep_request,ystep_request,zstep_request);

  enum Gv3Color { GV3_NONE, GV3_X, GV3_Y, GV3_Z, GV3_MAG, GV3_XYANGLE,
                  GV3_DIV};
  Gv3Color cq=GV3_NONE;
  if(colorquantity!=NULL) {
    if(strcmp("x",colorquantity)==0)             cq=GV3_X;
    else if(strcmp("y",colorquantity)==0)        cq=GV3_Y;
    else if(strcmp("z",colorquantity)==0)        cq=GV3_Z;
    else if(strcmp("mag",colorquantity)==0)      cq=GV3_MAG;
    else if(strcmp("xy-angle",colorquantity)==0) cq=GV3_XYANGLE;
    else if(strcmp("div",colorquantity)==0)      cq=GV3_DIV;
  }

  display_list.Clear();
  INT4m isize,jsize,ksize;
  GetDimens(isize,jsize,ksize);
  if(isize<1 || jsize<1 || ksize<1)
    return display_list.GetSize(); // Empty grid
  Nb_Vec3<REAL4> srangemin,srangemax;
  Nb_Vec3<REAL8> rangemin,rangemax;
  dsprange.GetExtremes(srangemin,srangemax);
  Convert(srangemin,rangemin);
  Convert(srangemax,rangemax);

  // Convert dsprange & step requests from problem coords to nodal coords
  rangemin=ProblemToNodalCoords(rangemin);
  rangemax=ProblemToNodalCoords(rangemax);
  if(coords_step.x<0) swap(rangemin.x,rangemax.x); // Adjust for ordering
  if(coords_step.y<0) swap(rangemin.y,rangemax.y); // across coord transform.
  if(coords_step.z<0) swap(rangemin.z,rangemax.z);

  REAL8m nxstep=xstep_request/fabs(coords_step.x);
  REAL8m nystep=ystep_request/fabs(coords_step.y);
  REAL8m nzstep=zstep_request/fabs(coords_step.z);

  INT4m ifirst,ilast,jfirst,jlast,kfirst,klast;
  ifirst=INT4m(ceil(rangemin.x)); ilast=INT4m(floor(rangemax.x))+1;
  jfirst=INT4m(ceil(rangemin.y)); jlast=INT4m(floor(rangemax.y))+1;
  kfirst=INT4m(ceil(rangemin.z)); klast=INT4m(floor(rangemax.z))+1;

  if(ifirst<0 || ilast>isize || jfirst<0 || jlast>jsize ||
     kfirst<0 || klast>ksize) {
#if DEBUGLVL > 999
    NonFatalError(STDDOC,"Range box out-of-bounds: (%d,%d,%d) "
		  "to (%d,%d,%d), not in (0,0,0) to (%d,%d,%d)",
		  ifirst,jfirst,kfirst,ilast,jlast,klast,
		  isize,jsize,ksize);
#endif
    if(ifirst<0) ifirst=0; if(ilast>isize) ilast=isize;
    if(jfirst<0) jfirst=0; if(jlast>jsize) jlast=jsize;
    if(kfirst<0) kfirst=0; if(klast>ksize) klast=ksize;
  }

  // Calculate step sizes
  // Note: Non-integer step size produce ugly banding in output sampling
  nxstep=OC_ROUND(nxstep);
  nystep=OC_ROUND(nystep);
  nzstep=OC_ROUND(nzstep);

  REAL8m max_step=REAL4m(OC_MAX(OC_MAX(isize,jsize),ksize));
  if(nxstep>max_step) nxstep=max_step;
  if(nystep>max_step) nystep=max_step;
  if(nzstep>max_step) nzstep=max_step;

  INT4m istep=INT4m(nxstep);
  INT4m jstep=INT4m(nystep);
  INT4m kstep=INT4m(nzstep);

  if(istep<1) istep=1;  if(jstep<1) jstep=1;  if(kstep<1) kstep=1;

  // Determine offsets
  INT4m ioff=((ilast-ifirst-1)%istep)/2;
  INT4m joff=((jlast-jfirst-1)%jstep)/2;
  INT4m koff=((klast-kfirst-1)%kstep)/2;

  INT4m i,j,k;
  INT4m istart,jstart,kstart;
  Nb_Vec3<REAL4> position;  // Position, in "Problem" coords
  REAL8m xstart,ystart,zstart,xstep,ystep,zstep;

  istart=ifirst+ioff;
  xstart=coords_base.x+coords_step.x*REAL8m(istart);
  xstep=coords_step.x*REAL8m(istep);

  jstart=jfirst+joff;
  ystart=coords_base.y+coords_step.y*REAL8m(jstart);
  ystep=coords_step.y*REAL8m(jstep);

  kstart=kfirst+koff;
  zstart=coords_base.z+coords_step.z*REAL8m(kstart);
  zstep=coords_step.z*REAL8m(kstep);


  REAL8m magsq,minmagsq;
  if(MinMagHint>MaxMagHint) {
    NonFatalError(STDDOC,"MinMagHint (%g) > MaxMagHintValue (%g)",
		  MinMagHint,MaxMagHint);
    return display_list.GetSize(); // Empty grid
  }
  if(MaxMagHint>0.0) {
    minmagsq=MinMagHint/MaxMagHint;
    minmagsq*=minmagsq;
  } else {
    minmagsq=1e-16;  // Pretend MaxMagHint=1.0, and MinMagHint=1e-8
  }

  Nb_Vec3<REAL8> wv;
  Nb_Vec3<REAL4> v;
  REAL8m wvscale=GetRelativeValueScale();
  if(MaxMagHint>1.0 && wvscale>FLT_MAX/MaxMagHint) {
    NonFatalError(STDDOC,"Value scale range error (too big): %g",wvscale);
    return display_list.GetSize(); // Empty grid
  }
  wvscale*=MaxMagHint;
  if(wvscale==0.0) wvscale=1.0; // Safety
  REAL8m wvtest= ( fabs(wvscale)<1.0 ? 1.0 : wvscale );
  wvtest=1.0/wvtest;
  REAL8m wvmult=1.0/wvscale;

  REAL8m xdivscale=1.;  // Scaling for divergence calculation
  if(xstep!=0.) { xdivscale=ystep/xstep; }

  REAL8m maxdiv=1.0;  // Dummy init value
  if(cq==GV3_DIV) {
    if(xdivscale>1+(FLT_MAX*wvtest/2.0)) {
      NonFatalError(STDDOC,"Divergence scaling range error (too big): %g",
                    xdivscale);
      return display_list.GetSize(); // Empty grid
    } else {
      maxdiv=2*wvscale*(1.+xdivscale); // Nominal max divergence
    }
  }

  REAL4m shade;
  REAL8m tempshade;

  for(k=kstart,position.z=zstart;k<klast;k+=kstep,position.z+=zstep) {
    for(j=jstart,position.y=ystart;j<jlast;j+=jstep,position.y+=ystep) {
      for(i=istart,position.x=xstart;i<ilast;i+=istep,position.x+=xstep) {
	wv=GridVec(i,j,k);
	magsq=wv.MagSq();
	if(magsq>FLT_MAX*wvtest*wvtest) {
	  // Vector will overflow floating point range; resize
	  REAL8m corrat = (FLT_MAX*wvtest*wvtest)/magsq;
	  magsq*=corrat;
	  wv*=1.0/sqrt(corrat);
	} else {
	  magsq*=wvmult*wvmult;
	  wv*=wvmult; // User specified data value scaling
	}
	if(UseMagHints) {
	  if(magsq<minmagsq) continue;  // Skip tiny vectors
	  if(magsq>1.0) { // Cut back (trim) overly large vectors
	    wv*=1.0/sqrt(magsq);
            magsq=1.0;  // Safety
	  }
	}
	Convert(wv,v);  // Narrow from REAL8 to REAL4
	shade=0.5;
	switch(cq) {
	case GV3_X:
	  if(1.0>fabs(v.x)) shade=(1.+v.x)/2.;
	  else if(v.x>0)    shade=1.0;
	  else              shade=0.0;
	  break;
	case GV3_Y:
	  if(1.0>fabs(v.y)) shade=(1.+v.y)/2.;
	  else if(v.y>0)    shade=1.0;
	  else              shade=0.0;
	  break;
	case GV3_Z:
	  if(1.0>fabs(v.z)) shade=(1.+v.z)/2.;
	  else if(v.z>0)    shade=1.0;
	  else              shade=0.0;
	  break;
	case GV3_MAG:
	  magsq=v.MagSq();
	  if(magsq>=1.0) shade=1.0;
	  else           shade=sqrt(magsq);
	  break;
	case GV3_XYANGLE:
	  shade = atan2(v.y,v.x)*(1./(2*PI));
	  if(shade<0) shade+=1.0;
	  break;
        case GV3_DIV:
	  shade=0.;
	  if(i+istep<isize) shade+=GridVec(i+istep,j,k).x;
	  if(i-istep>=0)    shade-=GridVec(i-istep,j,k).x;
	  shade*=xdivscale;
	  if(j+jstep<jsize) shade+=GridVec(i,j+jstep,k).y;
	  if(j-jstep>=0)    shade-=GridVec(i,j-jstep,k).y;
          tempshade=shade*wvmult;
	  if(maxdiv>fabs(tempshade)) shade=(1.+tempshade/maxdiv)/2.;
	  else if(tempshade>0)       shade=1.0;
	  else                       shade=0.0;
	  break;
	default:
	  break;
	}
	display_list.Append(Vf_DisplayVector(position,v,shade));
      }
    }
  }

  // Export actual step sizes, in problem units
  xstep_request=nxstep*fabs(coords_step.x);
  ystep_request=nystep*fabs(coords_step.y);
  zstep_request=nzstep*fabs(coords_step.z);

  return display_list.GetSize();
#undef MEMBERNAME
}

int Vf_GridVec3f::GetFirstPt(Vf_Mesh_Index* &key_export,
			     Nb_LocatedVector<REAL8> &lv) const
{
  key_export=(Vf_Mesh_Index*)(new Vf_GridVec3f_Index);
  if(GetSize()<=0) return 0;

  lv.location=GetBasePoint();
  lv.value=GridVec(0,0,0);
  return 1;
}

int Vf_GridVec3f::GetNextPt(Vf_Mesh_Index* &key_import,
			    Nb_LocatedVector<REAL8> &lv) const
{
  Vf_GridVec3f_Index *key=(Vf_GridVec3f_Index *)key_import;
  int isize,jsize,ksize;
  GetDimens(isize,jsize,ksize);
  int i=key->i,j=key->j,k=key->k;
  ++i;
  if(i>=isize) { i=0; ++j; }
  if(j>=jsize) { j=0; ++k; }
  if(k>=ksize) return 0;
  lv.location=GetBasePoint();
  Nb_Vec3<REAL8> step=GetGridStep();
  step.x*=i;
  step.y*=j;
  step.z*=k;
  lv.location+=step;
  lv.value=GridVec(i,j,k);
  key->i=i; key->j=j; key->k=k;
  return 1;
}

//////////////////////////////////////////////////////////////////////////
// General (non-regular) 3D mesh of Nb_Vec3<REAL4>'s
const ClassDoc Vf_GeneralMesh3f::class_doc("Vf_GeneralMesh3f",
		    "Michael J. Donahue (michael.donahue@nist.gov)",
		    "1.0.0","1-Sep-1997");

// Generic Mesh functions /////////////////////

INT4m
Vf_GeneralMesh3f::GetPreciseRange(Nb_BoundingBox<REAL8> &myrange) const
{
  myrange=range;
  return 0;
}

INT4m
Vf_GeneralMesh3f::GetPreciseDataRange(Nb_BoundingBox<REAL8> &myrange) const
{
  myrange=data_range;
  return 0;
}

INT4m
Vf_GeneralMesh3f::GetBoundaryList(Nb_List< Nb_Vec3<REAL4> >
				  &boundary_list) const
{
  Nb_Vec3<REAL4> newpt;
  const Nb_Vec3<REAL8> *vptr;
  Nb_List_Index< Nb_Vec3<REAL8> > key;
  boundary_list.Clear();
  for(vptr=boundary.GetFirst(key);vptr!=NULL;vptr=boundary.GetNext(key)) {
    newpt.Set(vptr->x,vptr->y,vptr->z);
    boundary_list.Append(newpt);
  }
  return 0;
}

INT4m Vf_GeneralMesh3f::ColorQuantityTypes(Nb_List<Nb_DString> &types) const
{
  types.Clear();
  types.Append(Nb_DString("x"));
  types.Append(Nb_DString("y"));
  types.Append(Nb_DString("z"));
  types.Append(Nb_DString("mag"));
  types.Append(Nb_DString("xy-angle"));
  types.Append(Nb_DString("none"));
  return types.GetSize();
}

REAL4m Vf_GeneralMesh3f::GetVecShade(const char* colorquantity,
                                     const Nb_Vec3<REAL4>& v) const
{
  REAL4m shade = 0.5;
  if(strcmp("xy-angle",colorquantity)==0) {
    shade = atan2(v.y,v.x)*(1./(2*PI));
    if(shade<0) shade+=1.0;
  } else {
    REAL8m mag=v.Mag();
    if(UseMagHints) {
      if(mag>MaxMagHint) mag=MaxMagHint;
      if(mag<MinMagHint) mag=1.0;
    }
    REAL8m val=0.0;
    if(strcmp("x",colorquantity)==0)        val = v.x;
    else if(strcmp("y",colorquantity)==0)   val = v.y;
    else if(strcmp("z",colorquantity)==0)   val = v.z;
    else                                    return -1;
    if(fabs(val)<mag) shade = (1.+val/mag)/2.;
    else if(val>0.)   shade = 1.0;
    else              shade = 0.0;
  }
  return shade;
}


INT4m Vf_GeneralMesh3f::FindPreciseClosest(const Nb_Vec3<REAL8> &pos,
					   Nb_LocatedVector<REAL8> &lv)
{
  INT4m dummy;
  Nb_Vec3<REAL8> work_pos=pos;
  Nb_LocatedVector<REAL8> *lvp;
  data_range.MoveInto(work_pos);
  INT4m errorcode=mesh.GetClosest2D(work_pos,lvp,dummy);
  if(errorcode==0) lv=*lvp;
  return errorcode;
}

REAL8m
Vf_GeneralMesh3f::GetMaxMag() const
{
  REAL8m magsq,maxmagsq=0.0;
  Vf_BoxList_Index key;
  const Nb_LocatedVector<REAL8> *lvp=mesh.GetFirst(key);
  while(lvp!=0) {
    magsq=lvp->value.MagSq();
    if(magsq>maxmagsq) maxmagsq=magsq;
    lvp=mesh.GetNext(key);
  }
  return sqrt(maxmagsq);
}

INT4m
Vf_GeneralMesh3f::GetDisplayList(REAL4m &xstep_request,
                                 REAL4m &ystep_request,
                                 REAL4m & /* zstep_request */,
                                 const Nb_BoundingBox<REAL4> & range_request,
                                 const char *colorquantity,
                                 Nb_List<Vf_DisplayVector> &display_list)
{
#define MEMBERNAME "GetDisplayList"
  enum Gm3Color { GM3_NONE, GM3_X, GM3_Y, GM3_Z, GM3_MAG, GM3_XYANGLE };
  Gm3Color cq=GM3_NONE;
  if(colorquantity!=NULL) {
    if(strcmp("x",colorquantity)==0)      cq=GM3_X;
    else if(strcmp("y",colorquantity)==0) cq=GM3_Y;
    else if(strcmp("z",colorquantity)==0) cq=GM3_Z;
    else if(strcmp("mag",colorquantity)==0)      cq=GM3_MAG;
    else if(strcmp("xy-angle",colorquantity)==0) cq=GM3_XYANGLE;
  }
  REAL4m shade;

  Nb_BoundingBox<REAL8> work_range(data_range);


  Nb_Vec3<REAL4> sminpt,smaxpt;
  Nb_Vec3<REAL8> minpt,maxpt;
  range_request.GetExtremes(sminpt,smaxpt);
  Convert(sminpt,minpt);
  Convert(smaxpt,maxpt);
  work_range.ExpandWith(minpt);
  work_range.ExpandWith(maxpt);
  work_range.GetExtremes(minpt,maxpt);

  REAL8m magsq,minmagsq;
  if(MinMagHint>MaxMagHint) {
    NonFatalError(STDDOC,"MinMagHint (%g) > MaxMagHintValue (%g)",
		  MinMagHint,MaxMagHint);
    return display_list.GetSize(); // Empty grid
  }

  if(MaxMagHint>0.0) {
    minmagsq=MinMagHint/MaxMagHint;
    minmagsq*=minmagsq;
  } else {
    minmagsq=1e-16;  // Pretend MaxMagHint=1.0, and MinMagHint=1e-8
  }

  REAL8m wvscale=GetRelativeValueScale();
  if(MaxMagHint>1.0 && wvscale>FLT_MAX/MaxMagHint) {
    NonFatalError(STDDOC,"Value scale range error (too big): %g",wvscale);
    return display_list.GetSize(); // Empty grid
  }
  wvscale*=MaxMagHint;
  if(wvscale==0.0) wvscale=1.0; // Safety
  REAL8m wvtest= ( fabs(wvscale)<1.0 ? 1.0 : wvscale );
  wvtest=1.0/wvtest;
  REAL8m wvmult=1.0/wvscale;

  display_list.Clear();
  mesh.ClearSelectCounts();
  if((minpt.x+xstep_request)==minpt.x ||
     (maxpt.x+xstep_request)==maxpt.x ||
     (minpt.y+ystep_request)==minpt.y ||
     (maxpt.y+ystep_request)==maxpt.y   ) {
    // Step request is 0 or too small; return whole list
    Nb_LocatedVector<REAL8> wlv;
    Nb_LocatedVector<REAL4> lv;
    mesh.ResetWholeAccess();
    while(mesh.GetWholeNext(wlv)==0) {

	// NB: There is a parallel branch below corresponding to the
	//     sub-sampled case

	shade=0.5;
	magsq=wlv.value.MagSq();
	if(magsq>FLT_MAX*wvtest*wvtest) {
	  // Vector will overflow floating point range; resize
	  REAL8m corrat = (FLT_MAX*wvtest*wvtest)/magsq;
	  magsq*=corrat;
	  wlv.value*=1.0/sqrt(corrat);
	} else {
	  magsq*=wvmult*wvmult;
	  wlv.value*=wvmult; // User specified data value scaling
	}
	if(UseMagHints) {
	  if(magsq<minmagsq) continue;  // Skip tiny vectors
	  if(magsq>1.0) { // Cut back (trim) overly large vectors
	    wlv.value*=1.0/sqrt(magsq);
	    magsq=1.0;
	  }
	}
	Convert(wlv.location,lv.location);
	Convert(wlv.value,lv.value);

	double val;
	switch(cq) {
	case GM3_X:
	  val=lv.value.x;
	  if(1.0>fabs(val)) shade=(1.+val)/2.;
	  else if(val>0)    shade=1.0;
	  else              shade=0.0;
	  break;
	case GM3_Y:
	  val=lv.value.y;
	  if(1.0>fabs(val)) shade=(1.+val)/2.;
	  else if(val>0)    shade=1.0;
	  else              shade=0.0;
	  break;
	case GM3_Z:
	  val=lv.value.z;
	  if(1.0>fabs(val)) shade=(1.+val)/2.;
	  else if(val>0)    shade=1.0;
	  else              shade=0.0;
	  break;
	case GM3_MAG:
	  if(magsq>=1.0) shade=1.0;
	  else           shade=sqrt(magsq);
	  break;
	case GM3_XYANGLE:
	  shade = atan2(lv.value.y,lv.value.x)*(1./(2*PI));
	  if(shade<0) shade+=1.0;
	  break;
	default:
	  break;
	}
	display_list.Append(Vf_DisplayVector(lv.location,lv.value,shade));
    }
  }
  else {
    // Otherwise, center a maximal regular grid, with lattice
    // steps xstep_request, ystep_request, inside working mesh box
    // region "work_range", and then use mesh.GetClosest2D to collect
    // samples from the mesh closest to the overlaid regular grid.
    const REAL8m fudge=1.001;  // Fudge factor, used to help
    /// against display "banding" caused by rounding.

    REAL8m xextent=maxpt.x-minpt.x;
    REAL8m xstep=fabs(xstep_request);
    INT4m  ixcount=int(floor((xextent*fudge)/xstep))+1;
    REAL8m xmargin=(xextent-xstep*(ixcount-1))/(2.*fudge);

    REAL8m yextent=maxpt.y-minpt.y;
    REAL8m ystep=fabs(ystep_request);
    INT4m  iycount=int(floor((yextent*fudge)/ystep))+1;
    REAL8m ymargin=(yextent-ystep*(iycount-1))/(2.*fudge);

    // If margin is close to an integral multiple of step_hints,
    // then round it to the multiple.  This should help remove
    // some display "banding" when sampling from grids that are
    // in actual fact regular.
    if(step_hints.x>0) {
      REAL8m frac=fmod(xmargin,step_hints.x);
      if(fabs(frac)<0.1) xmargin-=frac;
    }
    if(step_hints.y>0) {
      REAL8m frac=fmod(ymargin,step_hints.y);
      if(fabs(frac)<0.1) ymargin-=frac;
    }
    int ix,iy; REAL8m x,y;
    INT4m select_count;
    Nb_LocatedVector<REAL4> lv;
    Nb_LocatedVector<REAL8> wlv;
    Nb_LocatedVector<REAL8> *wlvp;
    Nb_Vec3<REAL8> pos(0.,0.,0.);
    for(ix=0, x=minpt.x+xmargin; ix<ixcount; ix++, x+=xstep) {
      pos.x=x;
      for(iy=0, y=minpt.y+ymargin; iy<iycount; iy++, y+=ystep) {
	// NB: There is a parallel branch above corresponding to the
	//     whole list (not sub-sampled) case
	pos.y=y;
	mesh.GetClosest2D(pos,wlvp,select_count);
	if(select_count==1) {
	  shade=0.5;
	  wlv = (*wlvp);
	  magsq=wlv.value.MagSq();
	  if(magsq>FLT_MAX*wvtest*wvtest) {
	    // Vector will overflow floating point range; resize
	    REAL8m corrat = (FLT_MAX*wvtest*wvtest)/magsq;
	    magsq*=corrat;
	    wlv.value*=1.0/sqrt(corrat);
	  } else {
	    magsq*=wvmult*wvmult;
	    wlv.value*=wvmult; // User specified data value scaling
	  }
	  if(UseMagHints) {
	    if(magsq<minmagsq) continue;  // Skip tiny vectors
	    if(magsq>1.0) { // Cut back (trim) overly large vectors
	      wlv.value*=1.0/sqrt(magsq);
	      magsq=1.0;
	    }
	  }
	  Convert(wlv.location,lv.location);
	  Convert(wlv.value,lv.value);

	  double val;
	  switch(cq) {
	  case GM3_X:
	    val=lv.value.x;
	    if(1.0>fabs(val)) shade=(1.+val)/2.;
	    else if(val>0)    shade=1.0;
	    else              shade=0.0;
	    break;
	  case GM3_Y:
	    val=lv.value.y;
	    if(1.0>fabs(val)) shade=(1.+val)/2.;
	    else if(val>0)    shade=1.0;
	    else              shade=0.0;
	    break;
	  case GM3_Z:
	    val=lv.value.z;
	    if(1.0>fabs(val)) shade=(1.+val)/2.;
	    else if(val>0)    shade=1.0;
	    else              shade=0.0;
	    break;
	  case GM3_MAG:
	    if(magsq>=1.0) shade=1.0;
	    else           shade=sqrt(magsq);
	    break;
	  case GM3_XYANGLE:
	    shade = atan2(lv.value.y,lv.value.x)*(1./(2*PI));
	    if(shade<0) shade+=1.0;
	    break;
	  default:
	    break;
	  }
	  display_list.Append(Vf_DisplayVector(lv.location,
                                               lv.value,shade));
	}
      }
    }
  }
  return display_list.GetSize();
#undef MEMBERNAME
}

void Vf_GeneralMesh3f::SetApproximateStepHints()
{ // We should have a sophisticated algorithm here that computes
  // and returns the average nearest neighbor distance, but for
  // now we'll just calculate some "average" cell dimensions.
  // NOTE: This functions overwrites any pre-existing value in
  //       member variable step_hints, which may override a file
  //       specified value.  A client may want to check to see
  //       if step_hints.x==0 before calling this routine.
#define MEMBERNAME "SetApproximateStepHints"
  step_hints.Set(0.,0.,0.);
  INT4m point_count=GetSize();
  if(point_count<1) return;  // Meaningless case
  REAL8m total_area=fabs(data_range.GetWidth()*data_range.GetHeight());
  /// fabs() for safety.

  REAL8m avecell_area=total_area/point_count;  // "Cells" here
  /// really mean "cells" for display purposes, which are centered
  /// about the sample nodes.

  step_hints.x=sqrt(avecell_area);
  step_hints.y=step_hints.x;
  step_hints.z=0.;
  return;
#undef MEMBERNAME
}

void Vf_GeneralMesh3f::SetBoundaryList(const Nb_List< Nb_Vec3<REAL8> >
                                       &boundary_list)
{
#define MEMBERNAME "SetBoundaryList"
  boundary=boundary_list;
  for(Nb_Vec3<REAL8> *vp=boundary.GetFirst();vp!=NULL;
      vp=boundary.GetNext()) {
    range.ExpandWith(*vp);
  }
#undef MEMBERNAME
}

REAL4m Vf_GeneralMesh3f::GetApproximateCellWidth() const
{ // Uses step_hints.  If step_hints.x==0, then first
  // calls SetApproximateStepHints.
  // NOTE: This function is "conceptually const" as opposed to
  //  "bitwise const"; cf. p76-78 of Meyer's book, "Effective C++."
  //  As a result, we need to do some ugly un-const casting.
  //  This is dangerous, since it looses compiler checks against
  //  changes to ::SetApproximateStepHints that may destroy
  //  conceptual const-ness of this function.
#define MEMBERNAME "GetApproximateCellWidth"
  Vf_GeneralMesh3f * const localThis = (Vf_GeneralMesh3f *)this;
  if(step_hints.x==0) localThis->SetApproximateStepHints();
  if(step_hints.x==0) return 1.0;  // Nonsense case
  return sqrt((OC_SQ(step_hints.x)+OC_SQ(step_hints.y))/2.);
  /// Using the diagonal instead of a simple average of step_hints.x
  /// and step_hints.y is more robust against non-square elements.
#undef MEMBERNAME
}

Nb_Vec3<REAL4> Vf_GeneralMesh3f::GetApproximateCellDimensions() const
{
  // See note in :::GetApproximateCellWidth() about
  // "conceptually const".
#define MEMBERNAME "GetApproximateCellDimensions"
  Vf_GeneralMesh3f * const localThis = (Vf_GeneralMesh3f *)this;
  if(step_hints.x==0) localThis->SetApproximateStepHints();
  if(step_hints.x==0) return Nb_Vec3<REAL4>(1.,1.,1.);  // Nonsense case
  Nb_Vec3<REAL4> temp;
  Convert(step_hints,temp);
  return temp;
#undef MEMBERNAME
}

// Vf_GeneralMesh3f specific functions
void Vf_GeneralMesh3f::ResetMesh()
{
#define MEMBERNAME "ResetMesh"
  mesh.ClearList();
  data_range.Reset();
  boundary.Clear();
  range.Reset();
  step_hints.Set(0.,0.,0.);
#undef MEMBERNAME
}

void Vf_GeneralMesh3f::ResetMesh(const char* newFilename)
{
#define MEMBERNAME "ResetMesh(const char*)"
  SetFilename(newFilename);
  ResetMesh();
#undef MEMBERNAME
}

void Vf_GeneralMesh3f::ResetMesh(const char* newFilename,
                                 const char* newTitle,
                                 const char* newDescription)
{
#define MEMBERNAME "ResetMesh(const char*,const char*)"
  SetTitle(newTitle);
  SetDescription(newDescription);
  ResetMesh(newFilename);
#undef MEMBERNAME
}

void Vf_GeneralMesh3f::AddPoint(const Nb_LocatedVector<REAL8>& lv)
{
#define MEMBERNAME "AddPoint"
  mesh.AddPoint(lv);
  data_range.ExpandWith(lv.location);
  range.ExpandWith(lv.location);
#undef MEMBERNAME
}

void Vf_GeneralMesh3f::SortPoints()
{
#define MEMBERNAME "SortPoints"
  // Guess at cell dimensions, if necessary
  if(step_hints.x==0) SetApproximateStepHints();

  // Make BoxList mesh bounding box big enough so that
  // refinement boxes will include all sampling points
  // we may ask of it.
  if(!range.IsEmpty()) {
    Nb_Vec3<REAL8> min,max;
    range.GetExtremes(min,max);
    mesh.ExpandBoxRegion(min.x,min.y,max.x,max.y);
  }
  mesh.InflateBoxRegion(1.2,1.2);   // Add a 10% margin on all sides

  // OK, now do refinement
  // mesh.Refine(6);
  mesh.Refine(1024,2.,15.);

  ClassDebugMessage(STDDOC,"Mesh refinement stats:\n"
	  "  Number of subboxes: %d\n"
	  "  Average number of points per box: %.2f\n"
	  "  Average box list length:  %.2f\n"
	  "  Memory waste: %d\n",
	  mesh.GetBoxCount(),mesh.GetAveInCount(),mesh.GetAveListCount(),
	  mesh.GetSpaceWaste());

#undef MEMBERNAME
}


int Vf_GeneralMesh3f::GetFirstPt(Vf_Mesh_Index* &key_export,
				 Nb_LocatedVector<REAL8> &lv) const
{
  Vf_GeneralMesh3f_Index *index=new Vf_GeneralMesh3f_Index;
  key_export=(Vf_Mesh_Index*)index;
  const Nb_LocatedVector<REAL8> *lvp=mesh.GetFirst(index->key);
  if(lvp==NULL) return 0;
  lv= *lvp;
  return 1;
}

int Vf_GeneralMesh3f::GetNextPt(Vf_Mesh_Index* &key_import,
				Nb_LocatedVector<REAL8> &lv) const
{
  Vf_GeneralMesh3f_Index *index=(Vf_GeneralMesh3f_Index *)key_import;
  const Nb_LocatedVector<REAL8> *lvp=mesh.GetNext(index->key);
  if(lvp==NULL) return 0;
  lv= *lvp;
  return 1;
}


